# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for the Pix3D dataset."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import tensorflow_datasets.public_api as tfds

from tensorflow_graphics.datasets import pix3d


class Pix3dTest(tfds.testing.DatasetBuilderTestCase):
  """Test Cases for Pix3D Dataset implementation."""
  DATASET_CLASS = pix3d.Pix3d
  SPLITS = {
      'train': 2,  # Number of fake train example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = ''
  EXAMPLE_DIR = os.path.join(os.path.dirname(__file__), 'fakes')
  MOCK_OUT_FORBIDDEN_OS_FUNCTIONS = False
  # SKIP_CHECKSUMS = True

  def setUp(self):  # pylint: disable=invalid-name
    super(Pix3dTest, self).setUp()
    self.builder.TRAIN_SPLIT_IDX = os.path.join(self.EXAMPLE_DIR,
                                                'pix3d_train.npy')
    self.builder.TEST_SPLIT_IDX = os.path.join(self.EXAMPLE_DIR,
                                               'pix3d_test.npy')


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Download, computes and stores the checksums."""

from absl import app
import tensorflow_datasets as tfds
from tensorflow_graphics.datasets.modelnet40 import ModelNet40


def main(_):
  config = tfds.download.DownloadConfig(register_checksums=True)
  modelnet40_builder = ModelNet40(data_dir="~/tensorflow_datasets")
  modelnet40_builder.download_and_prepare(download_config=config)


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Simple demo of modelnet40 dataset.

See: https://www.tensorflow.org/datasets/api_docs/python/tfds/load
"""
from absl import app

from tensorflow_graphics.datasets.modelnet40 import ModelNet40


def main(_):
  ds_train, info = ModelNet40.load(
      split="train", data_dir="~/tensorflow_dataset", with_info=True)

  for example in ds_train.take(1):
    points = example["points"]
    label = example["label"]

  # --- example accessing data
  print("points.shape=", points.shape)
  print("label.shape", label.shape)

  # --- example accessing info
  print("Example of string='{}' to ID#={}".format(
      "airplane", info.features["label"].str2int("airplane")))
  print("Example of ID#={} to string='{}'".format(
      12, info.features["label"].int2str(12)))


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow triangle utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def normal(v0, v1, v2, clockwise=False, normalize=True, name=None):
  """Computes face normals (triangles).

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    v0: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the first vertex of a triangle.
    v1: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the second vertex of a triangle.
    v2: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the third vertex of a triangle.
    clockwise: Winding order to determine front-facing triangles.
    normalize: A `bool` indicating whether output normals should be normalized
      by the function.
    name: A name for this op. Defaults to "triangle_normal".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
      a normalized vector.

  Raises:
    ValueError: If the shape of `v0`, `v1`, or `v2` is not supported.
  """
  with tf.compat.v1.name_scope(name, "triangle_normal", [v0, v1, v2]):
    v0 = tf.convert_to_tensor(value=v0)
    v1 = tf.convert_to_tensor(value=v1)
    v2 = tf.convert_to_tensor(value=v2)

    shape.check_static(tensor=v0, tensor_name="v0", has_dim_equals=(-1, 3))
    shape.check_static(tensor=v1, tensor_name="v1", has_dim_equals=(-1, 3))
    shape.check_static(tensor=v2, tensor_name="v2", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(v0, v1, v2), last_axes=-2, broadcast_compatible=True)

    normal_vector = vector.cross(v1 - v0, v2 - v0, axis=-1)
    normal_vector = asserts.assert_nonzero_norm(normal_vector)
    if not clockwise:
      normal_vector *= -1.0
    if normalize:
      return tf.nn.l2_normalize(normal_vector, axis=-1)
    return normal_vector


def area(v0, v1, v2, name=None):
  """Computes triangle areas.

    Note: Computed triangle area = 0.5 * | e1 x e2 | where e1 and e2 are edges
      of triangle. A degenerate triangle will return 0 area, whereas the normal
      for a degenerate triangle is not defined.


    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    v0: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the first vertex of a triangle.
    v1: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the second vertex of a triangle.
    v2: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the third vertex of a triangle.
    name: A name for this op. Defaults to "triangle_area".

  Returns:
    A tensor of shape `[A1, ..., An, 1]`, where the last dimension represents
      a normalized vector.
  """
  with tf.compat.v1.name_scope(name, "triangle_area", [v0, v1, v2]):
    v0 = tf.convert_to_tensor(value=v0)
    v1 = tf.convert_to_tensor(value=v1)
    v2 = tf.convert_to_tensor(value=v2)

    shape.check_static(tensor=v0, tensor_name="v0", has_dim_equals=(-1, 3))
    shape.check_static(tensor=v1, tensor_name="v1", has_dim_equals=(-1, 3))
    shape.check_static(tensor=v2, tensor_name="v2", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(v0, v1, v2), last_axes=-2, broadcast_compatible=True)

    normals = vector.cross(v1 - v0, v2 - v0, axis=-1)
    return 0.5 * tf.linalg.norm(tensor=normals, axis=-1, keepdims=True)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset Pipeline for mesh_segmentation_demo.ipynb.

 The shorthands used in parameter descriptions below are
    'B': Batch size.
    'E': Number of unique directed edges in a mesh.
    'V': Number of vertices in a mesh.
    'T': Number of triangles in a mesh.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.convolution import utils as conv_utils
from tensorflow_graphics.geometry.representation.mesh import utils as mesh_utils
from tensorflow_graphics.util import shape

DEFAULT_IO_PARAMS = {
    'batch_size': 8,
    'shuffle_buffer_size': 100,
    'is_training': True,
    'parallel_threads': 5,
    'mean_center': True,
    'shuffle': None,
    'repeat': None,
}


def adjacency_from_edges(edges, weights, num_edges, num_vertices):
  """Returns a batched sparse 1-ring adj tensor from edge list tensor.

  Args:
    edges: [B, E, 2] `int32` tensor of edges, possibly 0 padded.
    weights: [B, E] `float32` tensor of edge weights, possibly 0 padded.
    num_edges: [B] `int32` tensor of number of valid edges per batch sample.
    num_vertices: [B] `int32` tensor of number of valid vertices per batch
      sample.

  Returns:
    adj: A batched SparseTensor of weighted adjacency graph, of
      dense_shape [B, V, V] where V is max(num_vertices)
  """
  edges = tf.convert_to_tensor(value=edges)
  weights = tf.convert_to_tensor(value=weights)
  num_edges = tf.convert_to_tensor(value=num_edges)
  num_vertices = tf.convert_to_tensor(value=num_vertices)

  if not edges.dtype.is_integer:
    raise TypeError("'edges' must have an integer type.")
  if not num_edges.dtype.is_integer:
    raise TypeError("'num_edges' must have an integer type.")
  if not num_vertices.dtype.is_integer:
    raise TypeError("'num_vertices' must have an integer type.")
  if not weights.dtype.is_floating:
    raise TypeError("'weights' must have a floating type.")

  shape.check_static(tensor=edges, tensor_name='edges', has_rank=3)
  shape.check_static(tensor=weights, tensor_name='weights', has_rank=2)
  shape.check_static(tensor=num_edges, tensor_name='num_edges', has_rank=1)
  shape.check_static(
      tensor=num_vertices, tensor_name='num_vertices', has_rank=1)
  shape.compare_dimensions(
      tensors=(edges, weights, num_edges, num_vertices),
      tensor_names=('edges', 'weights', 'num_edges', 'num_vertices'),
      axes=(-3, -2, -1, -1))
  shape.compare_dimensions(
      tensors=(edges, weights),
      tensor_names=('edges', 'weights'),
      axes=(-2, -1))

  batch_size = tf.shape(input=edges)[0]
  max_num_vertices = tf.reduce_max(input_tensor=num_vertices)
  max_num_edges = tf.shape(input=edges)[1]
  batch_col = tf.reshape(tf.range(batch_size, dtype=edges.dtype), [-1, 1, 1])
  batch_col = tf.tile(batch_col, [1, max_num_edges, 1])
  batch_edges = tf.concat([batch_col, edges], axis=-1)

  indices, _ = conv_utils.flatten_batch_to_2d(batch_edges, sizes=num_edges)
  values, _ = conv_utils.flatten_batch_to_2d(
      tf.expand_dims(weights, -1), sizes=num_edges)
  values = tf.squeeze(values)
  adjacency = tf.SparseTensor(
      indices=tf.cast(indices, tf.int64),
      values=values,
      dense_shape=[batch_size, max_num_vertices, max_num_vertices])
  adjacency = tf.sparse.reorder(adjacency)
  return adjacency


def get_weighted_edges(faces, self_edges=True):
  r"""Gets unique edges and degree weights from a triangular mesh.

  The shorthands used below are:
      `T`: The number of triangles in the mesh.
      `E`: The number of unique directed edges in the mesh.

  Args:
    faces: A [T, 3] `int32` numpy.ndarray of triangle vertex indices.
    self_edges: A `bool` flag. If true, then for every vertex 'i' an edge
      [i, i] is added to edge list.
  Returns:
    edges: A  [E, 2] `int32` numpy.ndarray of directed edges.
    weights: A [E] `float32` numpy.ndarray denoting edge weights.

    The degree of a vertex is the number of edges incident on the vertex,
    including any self-edges. The weight for an edge $w_{ij}$ connecting vertex
    $v_i$ and vertex $v_j$ is defined as,
    $$
    w_{ij} = 1.0 / degree(v_i)
    \sum_{j} w_{ij} = 1
    $$
  """
  edges = mesh_utils.extract_unique_edges_from_triangular_mesh(
      faces, directed_edges=True).astype(np.int32)
  if self_edges:
    vertices = np.expand_dims(np.unique(edges[:, 0]), axis=1)
    self_edges = np.concatenate((vertices, vertices), axis=1)
    edges = np.unique(np.concatenate((edges, self_edges), axis=0), axis=0)
  weights = mesh_utils.get_degree_based_edge_weights(edges, dtype=np.float32)
  return edges, weights


def _tfrecords_to_dataset(tfrecords,
                          parallel_threads,
                          shuffle,
                          repeat,
                          sloppy,
                          max_readers=16):
  """Creates a TFRecordsDataset that iterates over filenames in parallel.

  Args:
    tfrecords: A list of tf.Data.TFRecords filenames.
    parallel_threads: The `int` number denoting number of parallel worker
      threads.
    shuffle: The `bool` flag denoting whether to shuffle the dataset.
    repeat: The `bool` flag denoting whether to repeat the dataset.
    sloppy: The `bool` flag denoting if elements are produced in deterministic
      order.
    max_readers: The `int` number denoting the maximum number of input tfrecords
      to interleave from in parallel.

  Returns:
    A tf.data.TFRecordDataset
  """

  total_tfrecords = sum([len(tf.io.gfile.glob(f)) for f in tfrecords])
  num_readers = min(total_tfrecords, max_readers)
  dataset = tf.data.Dataset.list_files(tfrecords, shuffle=shuffle)
  if repeat:
    dataset = dataset.repeat()
  return dataset.apply(
      tf.data.experimental.parallel_interleave(
          tf.data.TFRecordDataset,
          num_readers,
          sloppy=sloppy,
          buffer_output_elements=parallel_threads,
          prefetch_input_elements=parallel_threads))


def _parse_tfex_proto(example_proto):
  """Parses the tfexample proto to a raw mesh_data dictionary.

  Args:
    example_proto: A tf.Example proto storing the encoded mesh data.

  Returns:
    A mesh data dictionary with the following fields:
      'num_vertices': The `int64` number of vertices in mesh.
      'num_triangles': The `int64` number of triangles in mesh.
      'vertices': A serialized tensor of vertex positions.
      'triangles': A serialized tensor of triangle vertex indices.
      'labels': A serialized tensor of per vertex class labels.
  """
  feature_description = {
      'num_vertices': tf.io.FixedLenFeature([], tf.int64, default_value=0),
      'num_triangles': tf.io.FixedLenFeature([], tf.int64, default_value=0),
      'vertices': tf.io.FixedLenFeature([], tf.string, default_value=''),
      'triangles': tf.io.FixedLenFeature([], tf.string, default_value=''),
      'labels': tf.io.FixedLenFeature([], tf.string, default_value=''),
  }
  return tf.io.parse_single_example(
      serialized=example_proto, features=feature_description)


def _parse_mesh_data(mesh_data, mean_center=True):
  """Parses a raw mesh_data dictionary read from tf examples.

  Args:
    mesh_data: A mesh data dictionary with serialized data tensors,
      as output from _parse_tfex_proto()
    mean_center: If true, centers the mesh vertices to mean(vertices).
  Returns:
     A mesh data dictionary with following fields:
      'num_vertices': The `int32` number of vertices in mesh.
      'num_triangles': The `int32` number of triangles in mesh.
      'num_edges': The `int32` number of unique directed edges in mesh.
      'vertices': A [V, 3] `float32` of vertex positions.
      'triangles': A [T, 3] `int32` tensor of triangle vertex indices.
      'labels': A [V] `int32` tensor of per vertex class labels.
      'edges': A [E, 2] `int32` tensor of unique directed edges in mesh.
      'edge_weights': A [E] `float32` tensor of vertex degree based edge
        weights.
  """
  labels = tf.io.parse_tensor(mesh_data['labels'], tf.int32)
  vertices = tf.io.parse_tensor(mesh_data['vertices'], tf.float32)
  triangles = tf.io.parse_tensor(mesh_data['triangles'], tf.int32)
  if mean_center:
    vertices = vertices - tf.reduce_mean(
        input_tensor=vertices, axis=0, keepdims=True)

  edges, weights = tf.py_function(
      func=lambda t: get_weighted_edges(t.numpy()),
      inp=[triangles],
      Tout=[tf.int32, tf.float32])

  num_edges = tf.shape(input=edges)[0]
  num_vertices = tf.cast(mesh_data['num_vertices'], tf.int32)
  num_triangles = tf.cast(mesh_data['num_triangles'], tf.int32)
  mesh_data = dict(
      vertices=vertices,
      labels=labels,
      triangles=triangles,
      edges=edges,
      edge_weights=weights,
      num_triangles=num_triangles,
      num_vertices=num_vertices,
      num_edges=num_edges)
  return mesh_data


def create_dataset_from_tfrecords(tfrecords, params):
  """Creates a mesh dataset given a list of tf records filenames.

  Args:
    tfrecords: A list of TFRecords filenames.
    params: A dictionary of IO paramaters, see DEFAULT_IO_PARAMS.
  Returns:
    A tf.data.Dataset, with each element a dictionary of batched mesh data
      with following fields:
      'vertices': A [B, V, 3] `float32` tensor of vertex positions, possibly
        0-padded.
      'triangles': A [B, T, 3] `int32` tensor of triangle vertex indices,
        possibly 0-padded
      'labels': A [B, V] `int32` tensor of per vertex class labels, possibly
        0-padded
      'edges': A [B, E, 2] `int32` tensor of unique directed edges in mesh,
        possibly 0-padded
      'edge_weights': A [B, E] `float32` tensor of vertex degree based edge
        weights, possibly 0-padded.
      'num_edges': A [B] `int32` tensor of number of unique directed edges in
        each mesh in the batch.
      'num_vertices':  A [B] `int32` tensor of number of vertices in each mesh
        in the batch.
      'num_triangles': A [B] `int32` tensor of number of triangles in each mesh
        in the batch.
  """

  def _set_default_if_none(param, param_dict, default_val):
    if param not in param_dict:
      return default_val
    else:
      return default_val if param_dict[param] is None else param_dict[param]

  is_training = params['is_training']
  shuffle = _set_default_if_none('shuffle', params, is_training)
  repeat = _set_default_if_none('repeat', params, is_training)
  sloppy = _set_default_if_none('sloppy', params, is_training)

  if not isinstance(tfrecords, list):
    tfrecords = [tfrecords]
  dataset = _tfrecords_to_dataset(tfrecords, params['parallel_threads'],
                                  shuffle, repeat, sloppy)
  dataset = dataset.map(_parse_tfex_proto, tf.data.experimental.AUTOTUNE)
  dataset = dataset.map(
      lambda x: _parse_mesh_data(x, mean_center=params['mean_center']),
      tf.data.experimental.AUTOTUNE)
  if repeat:
    dataset = dataset.repeat()
  if shuffle:
    dataset = dataset.shuffle(params['shuffle_buffer_size'])
  return dataset.padded_batch(
      params['batch_size'],
      padded_shapes={
          'vertices': [None, 3],
          'labels': [None],
          'triangles': [None, 3],
          'edges': [None, 2],
          'edge_weights': [None],
          'num_edges': [],
          'num_vertices': [],
          'num_triangles': [],
      },
      drop_remainder=is_training)


def create_input_from_dataset(dataset_fn, files, io_params):
  """Creates input function given dataset generator and input files.

  Args:
    dataset_fn: A dataset generator function.
    files: A list of TFRecords filenames.
    io_params: A dictionary of IO paramaters, see DEFAULT_IO_PARAMS.
  Returns:
    features: A dictionary of mesh data training features.
    labels: A [B] `int32` tensor of per vertex class labels.
  """
  for k in DEFAULT_IO_PARAMS:
    io_params[k] = io_params[k] if k in io_params else DEFAULT_IO_PARAMS[k]

  dataset = dataset_fn(files, io_params)
  mesh_data = tf.compat.v1.data.make_one_shot_iterator(dataset).get_next()
  mesh_data['neighbors'] = adjacency_from_edges(mesh_data['edges'],
                                                mesh_data['edge_weights'],
                                                mesh_data['num_edges'],
                                                mesh_data['num_vertices'])

  max_num_verts = tf.reduce_max(input_tensor=mesh_data['num_vertices'])
  features = dict(
      vertices=tf.reshape(mesh_data['vertices'], [-1, max_num_verts, 3]),
      triangles=mesh_data['triangles'],
      neighbors=mesh_data['neighbors'],
      num_triangles=mesh_data['num_triangles'],
      num_vertices=mesh_data['num_vertices'])
  labels = mesh_data['labels']
  # Copy labels to features dictionary for estimator prediction mode.
  if not io_params['is_training']:
    features['labels'] = mesh_data['labels']
  return features, labels



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for utils."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.geometry.representation.mesh import utils
from tensorflow_graphics.util import test_case


class UtilsTest(test_case.TestCase):

  @parameterized.parameters(
      (np.array(((0, 1, 2),)), [[0, 1], [0, 2], [1, 2]]),
      (np.array(
          ((0, 1, 2), (0, 1, 3))), [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3]]),
  )
  def test_extract_undirected_edges_from_triangular_mesh_preset(
      self, test_inputs, test_outputs):
    """Tests that the output contain the expected edges."""
    edges = utils.extract_unique_edges_from_triangular_mesh(
        test_inputs, directed_edges=False)
    edges.sort(axis=1)  # Ensure edge tuple ordered by first vertex.
    self.assertEqual(sorted(edges.tolist()), test_outputs)

  @parameterized.parameters(
      (np.array(
          ((0, 1, 2),)), [[0, 1], [0, 2], [1, 0], [1, 2], [2, 0], [2, 1]]),
      (np.array(
          ((0, 1, 2), (0, 1, 3))), [[0, 1], [0, 2], [0, 3], [1, 0], [1, 2],
                                    [1, 3], [2, 0], [2, 1], [3, 0], [3, 1]]),
  )
  def test_extract_directed_edges_from_triangular_mesh_preset(
      self, test_inputs, test_outputs):
    """Tests that the output contain the expected edges."""
    edges = utils.extract_unique_edges_from_triangular_mesh(
        test_inputs, directed_edges=True)
    self.assertEqual(sorted(edges.tolist()), test_outputs)

  @parameterized.parameters(
      (1, "'faces' must be a numpy.ndarray."),
      (np.array((1,)), "must have a rank equal to 2"),
      (np.array((((1,),),)), "must have a rank equal to 2"),
      (np.array(((1,),)), "must have exactly 3 dimensions in the last axis"),
      (np.array(((1, 1),)), "must have exactly 3 dimensions in the last axis"),
      (np.array(
          ((1, 1, 1, 1),)), "must have exactly 3 dimensions in the last axis"),
  )
  def test_extract_edges_from_triangular_mesh_raised(
      self, invalid_input, error_msg):
    """Tests that the shape exceptions are properly raised."""
    with self.assertRaisesRegexp(ValueError, error_msg):
      utils.extract_unique_edges_from_triangular_mesh(invalid_input)

  @parameterized.parameters(
      (np.array(((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))),
       np.float16,
       [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
      (np.array(((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))),
       np.float32,
       [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
      (np.array(((0, 1), (0, 2), (0, 3), (1, 0), (1, 2), (1, 3),
                 (2, 0), (2, 1), (3, 0), (3, 1))),
       np.float64,
       [1.0 / 3, 1.0 / 3, 1.0 / 3, 1.0 / 3, 1.0 / 3, 1.0 / 3,
        0.5, 0.5, 0.5, 0.5]),
  )
  def test_get_degree_based_edge_weights_preset(
      self, test_inputs, test_dtype, test_outputs):
    """Tests that the output contain the expected edges."""
    weights = utils.get_degree_based_edge_weights(test_inputs, test_dtype)
    self.assertAllClose(weights.tolist(), test_outputs)

  @parameterized.parameters(
      (1, "'edges' must be a numpy.ndarray."),
      (np.array((1,)), "must have a rank equal to 2"),
      (np.array((((1,),),)), "must have a rank equal to 2"),
      (np.array(((1,),)), "must have exactly 2 dimensions in the last axis"),
      (np.array(
          ((1, 1, 1),)), "must have exactly 2 dimensions in the last axis"),
  )
  def test_get_degree_based_edge_weights_invalid_edges_raised(
      self, invalid_input, error_msg):
    """Tests that the shape exceptions are properly raised."""
    with self.assertRaisesRegexp(ValueError, error_msg):
      utils.get_degree_based_edge_weights(invalid_input)

  @parameterized.parameters(
      (np.bool, "must be a numpy float type"),
      (np.int, "must be a numpy float type"),
      (np.complex, "must be a numpy float type"),
      (np.uint, "must be a numpy float type"),
      (np.int16, "must be a numpy float type"),
  )
  def test_get_degree_based_edge_weights_dtype_raised(
      self, invalid_type, error_msg):
    """Tests that the shape exceptions are properly raised."""
    with self.assertRaisesRegexp(ValueError, error_msg):
      utils.get_degree_based_edge_weights(np.array(((1, 1),)), invalid_type)

if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements image pyramid functionalities.

More details about image pyramids can be found on [this page.]
(https://en.wikipedia.org/wiki/Pyramid_(image_processing))
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def _downsample(image, kernel):
  """Downsamples the image using a convolution with stride 2.

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    kernel: A tensor of shape `[H_k, W_k, C, C]`, where `H_k` and `W_k` are the
      height and width of the kernel.

  Returns:
    A tensor of shape `[B, H_d, W_d, C]`, where `H_d` and `W_d` are the height
    and width of the downsampled image.

  """
  return tf.nn.conv2d(
      input=image, filters=kernel, strides=[1, 2, 2, 1], padding="SAME")


def _binomial_kernel(num_channels, dtype=tf.float32):
  """Creates a 5x5 binomial kernel.

  Args:
    num_channels: The number of channels of the image to filter.
    dtype: The type of an element in the kernel.

  Returns:
    A tensor of shape `[5, 5, num_channels, num_channels]`.
  """
  kernel = np.array((1., 4., 6., 4., 1.), dtype=dtype.as_numpy_dtype)
  kernel = np.outer(kernel, kernel)
  kernel /= np.sum(kernel)
  kernel = kernel[:, :, np.newaxis, np.newaxis]
  return tf.constant(kernel, dtype=dtype) * tf.eye(num_channels, dtype=dtype)


def _build_pyramid(image, sampler, num_levels):
  """Creates the different levels of the pyramid.

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    sampler: A function to execute for each level (_upsample or _downsample).
    num_levels: The number of levels.

  Returns:
    A list containing `num_levels` tensors of shape `[B, H_i, W_i, C]`, where
    `H_i` and `W_i` are the height and width of the image for the level i.
  """
  kernel = _binomial_kernel(tf.shape(input=image)[3], dtype=image.dtype)
  levels = [image]
  for _ in range(num_levels):
    image = sampler(image, kernel)
    levels.append(image)
  return levels


def _split(image, kernel):
  """Splits the image into high and low frequencies.

  This is achieved by smoothing the input image and substracting the smoothed
  version from the input.

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    kernel: A tensor of shape `[H_k, W_k, C, C]`, where `H_k` and `W_k` are the
      height and width of the kernel.

  Returns:
    A tuple of two tensors of shape `[B, H, W, C]` and `[B, H_d, W_d, C]`, where
    the first one contains the high frequencies of the image and the second one
    the low frequencies. `H_d` and `W_d` are the height and width of the
    downsampled low frequency image.

  """
  low = _downsample(image, kernel)
  high = image - _upsample(low, kernel, tf.shape(input=image))
  return high, low


def _upsample(image, kernel, output_shape=None):
  """Upsamples the image using a transposed convolution with stride 2.

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    kernel: A tensor of shape `[H_k, W_k, C, C]`, where `H_k` and `W_k` are the
      height and width of the kernel.
    output_shape: The output shape.

  Returns:
    A tensor of shape `[B, H_u, W_u, C]`, where `H_u` and `W_u` are the height
    and width of the upsampled image.
  """
  if output_shape is None:
    output_shape = tf.shape(input=image)
    output_shape = (output_shape[0], output_shape[1] * 2, output_shape[2] * 2,
                    output_shape[3])
  return tf.nn.conv2d_transpose(
      image,
      kernel * 4.0,
      output_shape=output_shape,
      strides=[1, 2, 2, 1],
      padding="SAME")


def downsample(image, num_levels, name=None):
  """Generates the different levels of the pyramid (downsampling).

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    num_levels: The number of levels to generate.
    name: A name for this op that defaults to "pyramid_downsample".

  Returns:
    A list containing `num_levels` tensors of shape `[B, H_i, W_i, C]`, where
    `H_i` and `W_i` are the height and width of the downsampled image for the
    level i.

  Raises:
    ValueError: If the shape of `image` is not supported.
  """
  with tf.compat.v1.name_scope(name, "pyramid_downsample", [image]):
    image = tf.convert_to_tensor(value=image)

    shape.check_static(tensor=image, tensor_name="image", has_rank=4)

    return _build_pyramid(image, _downsample, num_levels)


def merge(levels, name=None):
  """Merges the different levels of the pyramid back to an image.

  Args:
    levels: A list containing tensors of shape `[B, H_i, W_i, C]`, where `B` is
      the batch size, H_i and W_i are the height and width of the image for the
      level i, and `C` the number of channels of the image.
    name: A name for this op that defaults to "pyramid_merge".

  Returns:
      A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.

  Raises:
    ValueError: If the shape of the elements of `levels` is not supported.
  """
  with tf.compat.v1.name_scope(name, "pyramid_merge", levels):
    levels = [tf.convert_to_tensor(value=level) for level in levels]

    for index, level in enumerate(levels):
      shape.check_static(
          tensor=level, tensor_name="level {}".format(index), has_rank=4)

    image = levels[-1]
    kernel = _binomial_kernel(tf.shape(input=image)[3], dtype=image.dtype)
    for level in reversed(levels[:-1]):
      image = _upsample(image, kernel, tf.shape(input=level)) + level
    return image


def split(image, num_levels, name=None):
  """Generates the different levels of the pyramid.

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    num_levels: The number of levels to generate.
    name: A name for this op that defaults to "pyramid_split".

  Returns:
    A list containing `num_levels` tensors of shape `[B, H_i, W_i, C]`, where
    `H_i` and `W_i` are the height and width of the image for the level i.

  Raises:
    ValueError: If the shape of `image` is not supported.
  """
  with tf.compat.v1.name_scope(name, "pyramid_split", [image]):
    image = tf.convert_to_tensor(value=image)

    shape.check_static(tensor=image, tensor_name="image", has_rank=4)

    kernel = _binomial_kernel(tf.shape(input=image)[3], dtype=image.dtype)
    low = image
    levels = []
    for _ in range(num_levels):
      high, low = _split(low, kernel)
      levels.append(high)
    levels.append(low)
    return levels


def upsample(image, num_levels, name=None):
  """Generates the different levels of the pyramid (upsampling).

  Args:
    image: A tensor of shape `[B, H, W, C]`, where `B` is the batch size, `H`
      the height of the image, `W` the width of the image, and `C` the number of
      channels of the image.
    num_levels: The number of levels to generate.
    name: A name for this op that defaults to "pyramid_upsample".

  Returns:
    A list containing `num_levels` tensors of shape `[B, H_i, W_i, C]`, where
    `H_i` and `W_i` are the height and width of the upsampled image for the
    level i.

  Raises:
    ValueError: If the shape of `image` is not supported.
  """
  with tf.compat.v1.name_scope(name, "pyramid_upsample", [image]):
    image = tf.convert_to_tensor(value=image)

    shape.check_static(tensor=image, tensor_name="image", has_rank=4)

    return _build_pyramid(image, _upsample, num_levels)

# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for emission absorption voxel rendering."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.rendering.voxels import emission_absorption
from tensorflow_graphics.rendering.voxels.tests import test_helpers
from tensorflow_graphics.util import test_case


class EmissionAbsorptionTest(test_case.TestCase):

  @parameterized.parameters(
      (0, (8, 16, 6, 1)),
      (1, (12, 8, 16, 6, 3)),
  )
  def test_render_shape_exception_not_raised(self, axis, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(emission_absorption.render,
                                        shape,
                                        axis=axis)

  @parameterized.parameters(
      ("must have a rank greater than 3", 2, (3,)),
      ("must have a rank greater than 3", 2, (16, 6, 3)),
      ("'axis' needs to be 0, 1 or 2", 5, (8, 16, 6, 1)),
  )
  def test_render_shape_exception_raised(self, error_msg, axis, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(emission_absorption.render,
                                    error_msg,
                                    shape,
                                    axis=axis)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_render_jacobian_random(self):
    """Tests the Jacobian of render."""
    voxels_init = test_helpers.generate_random_test_voxels_render()
    absorption_factor_init = np.float64(np.random.uniform(low=0.1, high=2.0))
    cell_size_init = np.float64(np.random.uniform(low=0.1, high=2.0))

    self.assert_jacobian_is_correct_fn(
        emission_absorption.render,
        [voxels_init, absorption_factor_init, cell_size_init],
        atol=1e-4)

  def test_render_preset(self):
    """Checks that render returns the expected value."""
    x_voxels_init, y_images_init = test_helpers.generate_preset_test_voxels_emission_absorption_render(
    )

    voxels = tf.convert_to_tensor(value=x_voxels_init)
    y_images = tf.convert_to_tensor(value=y_images_init)

    y = emission_absorption.render(voxels, absorption_factor=0.1, cell_size=0.1)

    self.assertAllClose(y_images, y)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""3D Pose feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow_datasets import features


class Pose(features.FeaturesDict):
  """`FeatureConnector` for 3d pose representations.

  During `_generate_examples`, the feature connector accepts as input any of:

    * `dict:` A dictionary containing the rotation and translation of the
    object (see output format below).

  Output:
    A dictionary containing:

    * 'R': A `float32` tensor with shape `[3, 3]` denoting the
    3D rotation matrix.
    * 't': A `float32` tensor with shape `[3,]` denoting the
    translation vector.

  """

  def __init__(self):
    super(Pose, self).__init__({
        'R': features.Tensor(shape=(3, 3), dtype=tf.float32),
        't': features.Tensor(shape=(3,), dtype=tf.float32),
    })

  def encode_example(self, example_dict):
    """Convert the given pose into a dict convertible to tf example."""

    if not all(key in example_dict for key in ['R', 't']):
      raise ValueError(
          f'Missing keys in provided dictionary! Expecting \'R\' and \'t\', '
          f'but {example_dict.keys()} were given.')

    return super(Pose, self).encode_example(example_dict)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query environment variable for documentation building."""
import os


def _import_tfg_docs():
  """Checks if __init__.py imports should be executed (for buildling docs)."""
  return os.getenv("TFG_DOC_IMPORTS", "0") == "1"


def enable_tfg_doc_imports():
  """Re-enables the imports in the __init__.py so that docs can be built."""
  os.environ["TFG_DOC_IMPORTS"] = "1"



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Shape utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

import six
import tensorflow as tf


def _broadcast_shape_helper(shape_x, shape_y):
  """Helper function for is_broadcast_compatible and broadcast_shape.

  Args:
    shape_x: A `TensorShape`.
    shape_y: A `TensorShape`.

  Returns:
    Returns None if the shapes are not broadcast compatible, or a list
    containing the broadcasted dimensions otherwise.
  """
  # To compute the broadcasted dimensions, we zip together shape_x and shape_y,
  # and pad with 1 to make them the same length.
  broadcasted_dims = reversed(
      list(
          six.moves.zip_longest(
              reversed(shape_x.dims),
              reversed(shape_y.dims),
              fillvalue=tf.compat.v1.Dimension(1))))
  # Next we combine the dimensions according to the numpy broadcasting rules.
  # http://docs.scipy.org/doc/numpy/user/basics.broadcasting.html
  return_dims = []
  for (dim_x, dim_y) in broadcasted_dims:
    if dim_x.value is None or dim_y.value is None:
      # One or both dimensions is unknown. If either dimension is greater than
      # 1, we assume that the program is correct, and the other dimension will
      # be broadcast to match it.
      if dim_x.value is not None and dim_x.value > 1:
        return_dims.append(dim_x)
      elif dim_y.value is not None and dim_y.value > 1:
        return_dims.append(dim_y)
      else:
        return_dims.append(None)
    elif dim_x.value == 1:
      # We will broadcast dim_x to dim_y.
      return_dims.append(dim_y)
    elif dim_y.value == 1:
      # We will broadcast dim_y to dim_x.
      return_dims.append(dim_x)
    elif dim_x.value == dim_y.value:
      # The dimensions are compatible, so output is the same size in that
      # dimension.
      return_dims.append(dim_x.merge_with(dim_y))
    else:
      return None
  return return_dims


def is_broadcast_compatible(shape_x, shape_y):
  """Returns True if `shape_x` and `shape_y` are broadcast compatible.

  Args:
    shape_x: A `TensorShape`.
    shape_y: A `TensorShape`.

  Returns:
    True if a shape exists that both `shape_x` and `shape_y` can be broadcasted
    to. False otherwise.
  """
  if shape_x.ndims is None or shape_y.ndims is None:
    return False
  return _broadcast_shape_helper(shape_x, shape_y) is not None


def get_broadcasted_shape(shape_x, shape_y):
  """Returns the common shape for broadcast compatible shapes.

  Args:
    shape_x: A `TensorShape`.
    shape_y: A `TensorShape`.

  Returns:
    Returns None if the shapes are not broadcast compatible, or a list
    containing the broadcasted dimensions otherwise.
  """
  if shape_x.ndims is None or shape_y.ndims is None:
    return None
  return _broadcast_shape_helper(shape_x, shape_y)


def _check_type(variable, variable_name, expected_type):
  """Helper function for checking that inputs are of expected types."""
  if isinstance(expected_type, (list, tuple)):
    expected_type_name = 'list or tuple'
  else:
    expected_type_name = expected_type.__name__
  if not isinstance(variable, expected_type):
    raise ValueError('{} must be of type {}, but it is {}'.format(
        variable_name, expected_type_name,
        type(variable).__name__))


def _fix_axis_dim_pairs(pairs, name):
  """Helper function to make `pairs` a list if needed."""
  if isinstance(pairs[0], int):
    pairs = [pairs]
  for pair in pairs:
    if len(pair) != 2:
      raise ValueError(
          '{} must consist of axis-value pairs, but found {}'.format(
              name, pair))
  return pairs


def _get_dim(tensor, axis):
  """Returns dimensionality of a tensor for a given axis."""
  return tf.compat.v1.dimension_value(tensor.shape[axis])


def check_static(tensor,
                 has_rank=None,
                 has_rank_greater_than=None,
                 has_rank_less_than=None,
                 has_dim_equals=None,
                 has_dim_greater_than=None,
                 has_dim_less_than=None,
                 tensor_name='tensor'):
  """Checks static shapes for rank and dimension constraints.

  This function can be used to check a tensor's shape for multiple rank and
  dimension constraints at the same time.

  Args:
    tensor: Any tensor with a static shape.
    has_rank: An int or `None`. If not `None`, the function checks if the rank
      of the `tensor` equals to `has_rank`.
    has_rank_greater_than: An int or `None`. If not `None`, the function checks
      if the rank of the `tensor` is greater than `has_rank_greater_than`.
    has_rank_less_than: An int or `None`. If not `None`, the function checks if
      the rank of the `tensor` is less than `has_rank_less_than`.
    has_dim_equals: Either a tuple or list containing a single pair of `int`s,
      or a list or tuple containing multiple such pairs. Each pair is in the
      form (`axis`, `dim`), which means the function should check if
      `tensor.shape[axis] == dim`.
    has_dim_greater_than: Either a tuple or list containing a single pair of
      `int`s, or a list or tuple containing multiple such pairs. Each pair is in
      the form (`axis`, `dim`), which means the function should check if
      `tensor.shape[axis] > dim`.
    has_dim_less_than: Either a tuple or list containing a single pair of
      `int`s, or a list or tuple containing multiple such pairs. Each pair is in
      the form (`axis`, `dim`), which means the function should check if
      `tensor.shape[axis] < dim`.
    tensor_name: A name for `tensor` to be used in the error message if one is
      thrown.

  Raises:
    ValueError: If any input is not of the expected types, or if one of the
      checks described above fails.
  """
  rank = tensor.shape.ndims

  def _raise_value_error_for_rank(variable, error_msg):
    raise ValueError(
        '{} must have a rank {} {}, but it has rank {} and shape {}'.format(
            tensor_name, error_msg, variable, rank, tensor.shape.as_list()))

  def _raise_value_error_for_dim(tensor_name, error_msg, axis, value):
    raise ValueError(
        '{} must have {} {} dimensions in axis {}, but it has shape {}'.format(
            tensor_name, error_msg, value, axis, tensor.shape.as_list()))

  if has_rank is not None:
    _check_type(has_rank, 'has_rank', int)
    if rank != has_rank:
      _raise_value_error_for_rank(has_rank, 'of')
  if has_rank_greater_than is not None:
    _check_type(has_rank_greater_than, 'has_rank_greater_than', int)
    if rank <= has_rank_greater_than:
      _raise_value_error_for_rank(has_rank_greater_than, 'greater than')
  if has_rank_less_than is not None:
    _check_type(has_rank_less_than, 'has_rank_less_than', int)
    if rank >= has_rank_less_than:
      _raise_value_error_for_rank(has_rank_less_than, 'less than')
  if has_dim_equals is not None:
    _check_type(has_dim_equals, 'has_dim_equals', (list, tuple))
    has_dim_equals = _fix_axis_dim_pairs(has_dim_equals, 'has_dim_equals')
    for axis, value in has_dim_equals:
      if _get_dim(tensor, axis) != value:
        _raise_value_error_for_dim(tensor_name, 'exactly', axis, value)
  if has_dim_greater_than is not None:
    _check_type(has_dim_greater_than, 'has_dim_greater_than', (list, tuple))
    has_dim_greater_than = _fix_axis_dim_pairs(has_dim_greater_than,
                                               'has_dim_greater_than')
    for axis, value in has_dim_greater_than:
      if not _get_dim(tensor, axis) > value:
        _raise_value_error_for_dim(tensor_name, 'greater than', axis, value)
  if has_dim_less_than is not None:
    _check_type(has_dim_less_than, 'has_dim_less_than', (list, tuple))
    has_dim_less_than = _fix_axis_dim_pairs(has_dim_less_than,
                                            'has_dim_less_than')
    for axis, value in has_dim_less_than:
      if not _get_dim(tensor, axis) < value:
        _raise_value_error_for_dim(tensor_name, 'less than', axis, value)


def _check_tensors(tensors, tensors_name):
  """Helper function to check the type and length of tensors."""
  _check_type(tensors, tensors_name, (list, tuple))
  if len(tensors) < 2:
    raise ValueError('At least 2 tensors are required.')


def _check_tensor_axis_lists(tensors, tensors_name, axes, axes_name):
  """Helper function to check that lengths of `tensors` and `axes` match."""
  _check_type(axes, axes_name, (list, tuple))
  if len(tensors) != len(axes):
    raise ValueError(
        '{} and {} must have the same length, but are {} and {}.'.format(
            tensors_name, axes_name, len(tensors), len(axes)))


def _fix_axes(tensors, axes, allow_negative):
  """Makes all axes positive and checks for out of bound errors."""
  axes = [
      axis + tensor.shape.ndims if axis < 0 else axis
      for tensor, axis in zip(tensors, axes)
  ]
  if not all(
      ((allow_negative or
        (not allow_negative and axis >= 0)) and axis < tensor.shape.ndims)
      for tensor, axis in zip(tensors, axes)):
    rank_axis_pairs = zip([tensor.shape.ndims for tensor in tensors], axes)
    raise ValueError(
        'Some axes are out of bounds. Given rank-axes pairs: {}'.format(
            [pair for pair in rank_axis_pairs]))
  return axes


def _give_default_names(list_of_objects, name):
  """Helper function to give default names to objects for error messages."""
  return [name + '_' + str(index) for index in range(len(list_of_objects))]


def _all_are_equal(list_of_objects):
  """Helper function to check if all the items in a list are the same."""
  if not list_of_objects:
    return True
  if isinstance(list_of_objects[0], list):
    list_of_objects = [tuple(obj) for obj in list_of_objects]
  return len(set(list_of_objects)) == 1


def _raise_error(tensor_names, batch_shapes):
  formatted_list = [(name, batch_shape)
                    for name, batch_shape in zip(tensor_names, batch_shapes)]
  raise ValueError(
      'Not all batch dimensions are identical: {}'.format(formatted_list))


def compare_batch_dimensions(tensors,
                             last_axes,
                             broadcast_compatible,
                             initial_axes=0,
                             tensor_names=None):
  """Compares batch dimensions for tensors with static shapes.

  Args:
    tensors: A list or tuple of tensors with static shapes to compare.
    last_axes: An `int` or a list or tuple of `int`s with the same length as
      `tensors`. If an `int`, it is assumed to be the same for all the tensors.
      Each entry should correspond to the last axis of the batch (with zero
      based indices). For instance, if there is only a single batch dimension,
      last axis should be `0`.
    broadcast_compatible: A 'bool', whether the batch shapes can be broadcast
      compatible in the numpy sense.
    initial_axes: An `int` or a list or tuple of `int`s with the same length as
      `tensors`. If an `int`, it is assumed to be the same for all the tensors.
      Each entry should correspond to the first axis of the batch (with zero
      based indices). Default value is `0`.
    tensor_names: Names of `tensors` to be used in the error message if one is
      thrown. If left as `None`, `tensor_i` is used.

  Raises:
    ValueError: If inputs have unexpected types, or if given axes are out of
      bounds, or if the check fails.
  """
  _check_tensors(tensors, 'tensors')
  if isinstance(initial_axes, int):
    initial_axes = [initial_axes] * len(tensors)
  if isinstance(last_axes, int):
    last_axes = [last_axes] * len(tensors)
  _check_tensor_axis_lists(tensors, 'tensors', initial_axes, 'initial_axes')
  _check_tensor_axis_lists(tensors, 'tensors', last_axes, 'last_axes')
  initial_axes = _fix_axes(tensors, initial_axes, allow_negative=True)
  last_axes = _fix_axes(tensors, last_axes, allow_negative=True)
  batch_shapes = [
      tensor.shape[init:last + 1]
      for tensor, init, last in zip(tensors, initial_axes, last_axes)
  ]
  if tensor_names is None:
    tensor_names = _give_default_names(tensors, 'tensor')
  if not broadcast_compatible:
    batch_ndims = [batch_shape.ndims for batch_shape in batch_shapes]
    batch_shapes = [batch_shape.as_list() for batch_shape in batch_shapes]
    if not _all_are_equal(batch_ndims):
      # If not all batch shapes have the same length, they cannot be identical.
      _raise_error(tensor_names, batch_shapes)
    for dims in zip(*batch_shapes):
      if _all_are_equal(dims):
        # Continue if all dimensions are None or have the same value.
        continue
      if None not in dims:
        # If all dimensions are known at this point, they are not identical.
        _raise_error(tensor_names, batch_shapes)
      # At this point dims must consist of both None's and int's.
      if len(set(dims)) != 2:
        # set(dims) should return (None, some_int).
        # Otherwise shapes are not identical.
        _raise_error(tensor_names, batch_shapes)
  else:
    if not all(
        is_broadcast_compatible(shape1, shape2)
        for shape1, shape2 in itertools.combinations(batch_shapes, 2)):
      raise ValueError(
          'Not all batch dimensions are broadcast-compatible: {}'.format([
              (name, batch_shape.as_list())
              for name, batch_shape in zip(tensor_names, batch_shapes)
          ]))


def compare_dimensions(tensors, axes, tensor_names=None):
  """Compares dimensions of tensors with static or dynamic shapes.

  Args:
    tensors: A list or tuple of tensors to compare.
    axes: An `int` or a list or tuple of `int`s with the same length as
      `tensors`. If an `int`, it is assumed to be the same for all the tensors.
      Each entry should correspond to the axis of the tensor being compared.
    tensor_names: Names of `tensors` to be used in the error message if one is
      thrown. If left as `None`, their `Tensor.name` fields are used instead.

  Raises:
    ValueError: If inputs have unexpected types, or if given axes are out of
      bounds, or if the check fails.
  """
  _check_tensors(tensors, 'tensors')
  if isinstance(axes, int):
    axes = [axes] * len(tensors)
  _check_tensor_axis_lists(tensors, 'tensors', axes, 'axes')
  axes = _fix_axes(tensors, axes, allow_negative=False)
  if tensor_names is None:
    tensor_names = _give_default_names(tensors, 'tensor')
  dimensions = [_get_dim(tensor, axis) for tensor, axis in zip(tensors, axes)]
  if not _all_are_equal(dimensions):
    raise ValueError('Tensors {} must have the same number of dimensions in '
                     'axes {}, but they are {}.'.format(
                         list(tensor_names), list(axes), list(dimensions)))


def is_static(tensor_shape):
  """Checks if the given tensor shape is static."""
  if isinstance(tensor_shape, (list, tuple)):
    return None not in tensor_shape
  else:
    return None not in tensor_shape.as_list()


# The util functions or classes are not exported.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Visualization in 3D of modelnet40 dataset.

See: https://www.tensorflow.org/datasets/api_docs/python/tfds/load
"""
from absl import app
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # pylint:disable=unused-import
from tensorflow_graphics.datasets.modelnet40 import ModelNet40


def main(_):
  ds_train, _ = ModelNet40.load(
      split="train", data_dir="~/tensorflow_dataset", with_info=True)

  for example in ds_train.take(1):
    points = example["points"]
    label = example["label"]

  fig = plt.figure()
  ax3 = fig.add_subplot(111, projection="3d")
  ax3.set_title("Example with label {}".format(label))
  scatter3 = lambda p, c="r", *args: ax3.scatter(p[:, 0], p[:, 1], p[:, 2], c)
  scatter3(points)


if __name__ == "__main__":
  app.run(main)
